use std::{
    sync::{Arc, RwLock},
    thread::JoinHandle,
    time::Duration,
    u64,
};

use chrono::Utc;
use rand::Rng;

pub enum Schedule {
    Every(Duration),
    At(chrono::DateTime<Utc>),
}

pub struct CronTask {
    f: Arc<Box<dyn Fn() + Send + Sync + 'static>>,
    schedule: Schedule,
    name: String,
    last_run: Option<chrono::DateTime<Utc>>,
}

impl CronTask {
    pub fn new<F: Fn() + Send + Sync + 'static>(name: &str, schedule: Schedule, f: F) -> Self {
        Self {
            f: Arc::new(Box::new(f)),
            schedule,
            name: name.to_string(),
            last_run: None,
        }
    }

    pub fn is_absolute(&self) -> bool {
        match self.schedule {
            Schedule::Every(_) => false,
            Schedule::At(_) => true,
        }
    }

    pub fn run(&mut self) -> JoinHandle<()> {
        log::info!("Starting cron task '{}'", self.name);
        self.last_run = Some(Utc::now());
        let f = Arc::clone(&self.f);
        std::thread::spawn(move || {
            f.as_ref()();
        })
    }

    pub fn wait_until(&mut self) -> Duration {
        match self.schedule {
            Schedule::Every(duration) => {
                let now = Utc::now();
                if let Some(last_exec) = self.last_run {
                    let since_then = (now - last_exec).to_std().unwrap();

                    duration.checked_sub(since_then).unwrap_or(Duration::ZERO)
                } else {
                    self.last_run = Some(Utc::now());
                    duration
                }
            }
            Schedule::At(date_time) => {
                if self.last_run.is_none() {
                    let now = Utc::now();
                    if let Ok(dur) = date_time.signed_duration_since(&now).to_std() {
                        dur
                    } else {
                        Duration::ZERO
                    }
                } else {
                    Duration::from_secs(u64::MAX)
                }
            }
        }
    }
}

pub struct Cron {
    tasks: RwLock<Vec<CronTask>>,
}

impl Cron {
    pub fn new() -> Self {
        Self {
            tasks: RwLock::new(Vec::new()),
        }
    }

    pub fn add_task<F: Fn() + Send + Sync + 'static>(&self, name: &str, schedule: Schedule, f: F) {
        self.tasks
            .write()
            .unwrap()
            .push(CronTask::new(name, schedule, f));
    }

    pub fn run_at<F: Fn() + Send + Sync + 'static>(&self, dt: chrono::DateTime<chrono::Utc>, f: F) {
        let name = format!("delayed_{}", rand::rng().random_range(1000..9999));
        self.tasks
            .write()
            .unwrap()
            .push(CronTask::new(&name, Schedule::At(dt), f));
    }

    pub fn run(&self) {
        loop {
            // init
            let mut last_wait = Duration::from_secs(u64::MAX);
            let mut last_task: Option<usize> = None;

            {
                // find next task
                let mut tasks = self.tasks.write().unwrap();
                for (i, task) in tasks.iter_mut().enumerate() {
                    let wait_time = task.wait_until();
                    if wait_time < last_wait {
                        last_wait = wait_time;
                        last_task = Some(i);
                    }
                }
            }

            if let Some(index) = last_task {
                // init
                let mut remove = false;
                let mut skip = false;

                // limit longest blocking time (5s)
                let real_wait = if last_wait.gt(&Duration::from_secs(5)) {
                    skip = true;
                    Duration::from_secs(5)
                } else {
                    last_wait
                };

                {
                    // logging
                    let tasks = self.tasks.read().unwrap();
                    log::debug!("Managing {} cron task(s)", tasks.len());

                    let task = tasks.get(index).unwrap();
                    if real_wait == last_wait {
                        log::debug!(
                            "Waiting for {real_wait:?} to start cron task '{}'",
                            task.name
                        );
                    } else {
                        log::debug!(
                            "Would wait for {last_wait:?} to start cron task '{}'. Waiting for {real_wait:?}",
                            task.name
                        );
                    }

                    // if somehow we wait indefinitely
                    if last_wait == Duration::from_secs(u64::MAX) {
                        log::warn!("Infinite wait time for cron");
                        continue;
                    }

                    // set remove flag for absolute time cron tasks
                    if task.is_absolute() {
                        log::info!(
                            "Removing task '{}' from cron because it will never run again",
                            task.name
                        );
                        remove = true;
                    }
                }

                // sleep until task
                std::thread::sleep(real_wait);

                // skip if we are still just sleeping
                if skip {
                    continue;
                }

                {
                    // run cron task
                    let mut tasks = self.tasks.write().unwrap();
                    let task = tasks.get_mut(index).unwrap();
                    let _ = task.run();
                }

                if remove {
                    {
                        // remove if requested
                        let mut tasks = self.tasks.write().unwrap();
                        log::info!("Removing cron task #{index}");
                        tasks.remove(index);
                    }
                }
            }
        }
    }
}
