// TODO : docs

// TODO : measure avg iteration time

use std::{sync::Arc, time::Duration};

use crate::{
    job::{JobDispatch, JobDispatcher},
    retry,
};

#[derive(Debug, Clone)]
pub struct FunctionContext<C, I, O> {
    pub context: C,
    pub initial: I,
    pub output: Option<O>,
    done: bool,
}

impl<C: Default, I, O> FunctionContext<C, I, O> {
    pub fn new(input: I) -> FunctionContext<C, I, O> {
        FunctionContext {
            context: C::default(),
            initial: input,
            output: None,
            done: false,
        }
    }
}

impl<C, I, O> FunctionContext<C, I, O> {
    pub fn state<F: Fn(&mut C)>(&mut self, f: F) {
        f(&mut self.context);
    }

    pub fn done(&mut self, output: O) {
        self.done = true;
        self.output = Some(output);
    }
}

pub struct RunningIteratedFunction<C, I, O> {
    output: std::marker::PhantomData<O>,
    context: FunctionContext<C, I, O>,
    f: Arc<Box<dyn Fn(FunctionContext<C, I, O>) -> FunctionContext<C, I, O>>>,
}

impl<C, I, O> RunningIteratedFunction<C, I, O> {
    pub fn next(mut self) -> Self {
        let new_ctx = self.f.as_ref()(self.context);
        self.context = new_ctx;
        return self;
    }

    pub fn return_value(&self) -> Option<&O> {
        self.context.output.as_ref()
    }

    pub fn take_return_value(self) -> O {
        self.context.output.unwrap()
    }

    pub fn is_done(&self) -> bool {
        self.context.done
    }
}

pub struct ThreadRunningIteratedFunction<C: Send + 'static, I: Send + 'static, O: Send + 'static>(
    JobDispatcher<IteratedFnQuery, IteratedFnOutput<C, I, O>>,
);

impl<C: Send + 'static, I: Send + 'static, O: Send + 'static>
    ThreadRunningIteratedFunction<C, I, O>
{
    pub fn start(&self) {
        let _ = self.0.try_send(IteratedFnQuery::Start);
    }

    pub fn pause(&self) {
        let _ = self.0.try_send(IteratedFnQuery::Pause);
    }

    pub fn stop(&self) -> FunctionContext<C, I, O> {
        match self.0.try_send(IteratedFnQuery::Stop) {
            Some(IteratedFnOutput::Context(function_context)) => function_context,
            _ => unreachable!(),
        }
    }

    pub fn try_output(&self) -> Option<O> {
        match self.0.send(IteratedFnQuery::GetOutput) {
            IteratedFnOutput::Out(out) => Some(out),
            _ => None,
        }
    }

    pub fn output(&self) -> O {
        retry(|| self.try_output())
    }
}

#[derive(Debug, Clone)]
pub enum IteratedFnQuery {
    Pause,
    Start,
    Stop,
    GetOutput,
}

pub enum IteratedFnOutput<C, I, O> {
    Out(O),
    Context(FunctionContext<C, I, O>),
    Ok,
}

pub struct IteratedFunction<C, I, O> {
    output: std::marker::PhantomData<O>,
    f: Arc<Box<dyn Fn(FunctionContext<C, I, O>) -> FunctionContext<C, I, O>>>,
}

impl<C: Default + Clone + Send, I: Clone + Send + 'static, O: Send + Clone>
    IteratedFunction<C, I, O>
{
    pub fn new<F: Fn(FunctionContext<C, I, O>) -> FunctionContext<C, I, O> + 'static>(
        f: F,
    ) -> Self {
        Self {
            f: Arc::new(Box::new(f)),
            output: std::marker::PhantomData,
        }
    }

    pub fn new_threaded<
        F: Fn(FunctionContext<C, I, O>) -> FunctionContext<C, I, O> + Send + 'static,
    >(
        f: F,
        input: I,
    ) -> ThreadRunningIteratedFunction<C, I, O> {
        Self::new_threaded_from_state(f, FunctionContext::new(input))
    }

    pub fn new_threaded_from_state<
        F: Fn(FunctionContext<C, I, O>) -> FunctionContext<C, I, O> + Send + 'static,
    >(
        f: F,
        context: FunctionContext<C, I, O>,
    ) -> ThreadRunningIteratedFunction<C, I, O> {
        let (dispatch, recv) = JobDispatcher::<IteratedFnQuery, IteratedFnOutput<C, I, O>>::new();

        let _ = std::thread::spawn(move || {
            let f = Self::new(f);
            let mut f = f.call_with_context(context);

            let mut counter = 0;
            let mut sleep = false;
            while !f.is_done() {
                if sleep {
                    std::thread::sleep(Duration::from_secs(3));
                }

                if counter == 5 || sleep {
                    if let Ok(request) = recv.recv_timeout(Duration::from_millis(300)) {
                        match request.param {
                            IteratedFnQuery::Pause => {
                                log::info!("Paused threaded iterative function");
                                sleep = true;
                            }
                            IteratedFnQuery::Start => {
                                log::info!("Restarted threaded iterative function");
                                sleep = false;
                            }
                            IteratedFnQuery::Stop => {
                                log::info!("Ending threaded iterative function");
                                request.done(IteratedFnOutput::Context(f.context));
                                return;
                            }
                            _ => {}
                        }

                        request.done(IteratedFnOutput::Ok);
                    }
                    counter = 0;
                }

                if !sleep {
                    f = f.next();
                }

                counter += 1;
            }

            if f.is_done() {
                while let Ok(request) = recv.recv() {
                    match request.param {
                        IteratedFnQuery::Stop => {
                            log::warn!("Function was asked to stop but was already done");
                            request.done(IteratedFnOutput::Context(f.context.clone()));
                        }
                        IteratedFnQuery::GetOutput => {
                            request.done(IteratedFnOutput::Out(f.context.output.clone().unwrap()));
                            break;
                        }
                        _ => {
                            request.done(IteratedFnOutput::Out(f.context.output.clone().unwrap()));
                        }
                    }
                }
            }
        });

        ThreadRunningIteratedFunction(dispatch)
    }

    pub fn call_with_context(
        &self,
        ctx: FunctionContext<C, I, O>,
    ) -> RunningIteratedFunction<C, I, O> {
        RunningIteratedFunction {
            output: std::marker::PhantomData,
            context: ctx,
            f: self.f.clone(),
        }
    }

    pub fn call(&self, input: I) -> RunningIteratedFunction<C, I, O> {
        RunningIteratedFunction {
            output: std::marker::PhantomData,
            context: FunctionContext::new(input),
            f: self.f.clone(),
        }
    }

    pub fn run_to_end(&self, input: I) -> O {
        let mut f = self.call(input);
        while !f.is_done() {
            f = f.next();
        }
        return f.take_return_value();
    }
}
